/**	\file Wsdl.cpp
 **	\date  2008-05-26
 **	\author grymse@alhem.net
**/
/*
Copyright (C) 2008  Anders Hedstrom

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#include "Wsdl.h"
#include <Sockets/XmlNode.h>
#include "WsdlException.h"
#include "Split.h"
#include <iostream>
#include "ComplexType.h"
#include "Application.h"
#include "SimpleType.h"
#include <Sockets/Debug.h>

#define DEB(x) x


Wsdl::Wsdl(Application& appl, Xml::XmlDocument& doc) : m_appl(appl), m_doc(doc)
{
	try
	{
		ParseElements();
	}
	catch (const Exception& e)
	{
		std::cout << "ParseElements() failed: " + e.ToString() << std::endl;
		abort();
	}
	m_xml_schema_prefix = GetNamespacePrefix("http://www.w3.org/2001/XMLSchema");
}


Wsdl::~Wsdl()
{
}


const std::string Wsdl::GetTargetNamespace() const
{
	Xml::XmlNode root(m_doc);
	return root.GetProperty("targetNamespace");
}


const std::string Wsdl::GetTargetNamespacePrefix() const
{
	Xml::XmlNode root(m_doc);
	return root.GetNsMapRe()[root.GetProperty("targetNamespace")];
}


const std::string Wsdl::GetNamespacePrefix(const std::string& href) const
{
	Xml::XmlNode root(m_doc);
	return root.GetNsMapRe()[href];
}


std::list<std::string> Wsdl::GetPorts() const
{
	std::list<std::string> vec;
	Xml::XmlNode n(m_doc, "service");
	if (n)
	{
		Xml::XmlNode n(n, "port");
		while (n)
		{
			vec.push_back(n.GetProperty("name"));
			++n;
		}
	}
	return vec;
}


const std::string Wsdl::GetPortname() const
{
	Xml::XmlNode n(m_doc, "portType");
	if (n)
		return n.GetProperty("name");
	throw WsdlException("No portType found in .wsdl document");
}


std::list<std::string> Wsdl::GetOperations(const std::string& portname) const
{
	std::list<std::string> vec;
	Xml::XmlNode n(m_doc, "portType");
	while (n)
	{
		if (n.GetProperty("name") == portname || portname.empty())
		{
			Xml::XmlNode n(n, "operation");
			while (n)
			{
				vec.push_back(n.GetProperty("name"));
				++n;
			}
		}
		++n;
	}
	return vec;
}


const std::string Wsdl::GetSoapAction(const std::string& portname, const std::string& op) const
{
	Xml::XmlNode n = GetBindingOp(portname, op);
	if (n)
	{
		Xml::XmlNode n(n, "operation");
		if (n)
		{
			return n.GetProperty("soapAction");
		}
	}
	throw WsdlException("SoapAction not found");
}


const std::string Wsdl::GetMessage(const std::string& portname, const std::string& op, const std::string& type) const
{
	Xml::XmlNode n(m_doc, "portType");
	while (n)
	{
		if (n.GetProperty("name") == portname || portname.empty())
		{
			Xml::XmlNode n(n, "operation");
			while (n)
			{
				if (n.GetProperty("name") == op)
				{
					Xml::XmlNode n(n, type);
					if (n)
					{
						return n.GetProperty("message");
					}
				}
				++n;
			}
		}
		++n;
	}
	throw WsdlException(type + " message not found for: " + portname + "/" + op);
}


Xml::XmlNode Wsdl::GetPortTypeOp(const std::string& portname, const std::string& op) const
{
	Xml::XmlNode n(m_doc, "portType");
	while (n)
	{
		if (n.GetProperty("name") == portname || portname.empty())
		{
			Xml::XmlNode n(n, "operation");
			while (n)
			{
				if (n.GetProperty("name") == op)
				{
					return n;
				}
				++n;
			}
		}
		++n;
	}
	throw WsdlException("portType/operation not found: " + portname + "/" + op);
}


Xml::XmlNode Wsdl::GetBindingOp(const std::string& portname, const std::string& op) const
{
	Xml::XmlNode n(m_doc, "binding");
	while (n)
	{
		if (n.GetProperty("name") == portname || portname.empty())
		{
			Xml::XmlNode n(n, "operation");
			while (n)
			{
				if (n.GetProperty("name") == op)
				{
					return n;
				}
				++n;
			}
		}
		++n;
	}
	throw WsdlException("binding/operation not found: " + portname + "/" + op);
}


Xml::XmlNode Wsdl::GetMessage(const std::string& msgname) const
{
	Xml::XmlNode n(m_doc, "message");
	while (n)
	{
		if (n.GetProperty("name") == Split(msgname).Type)
		{
			return n;
		}
		++n;
	}
	throw WsdlException("Message not found: " + msgname);
}


void Wsdl::ShowMessage(const std::string& msgname)
{
	Xml::XmlNode message = GetMessage(msgname);
	Xml::XmlNode n(message, "part");
	while (n)
	{
		std::string name;
		std::string type;
		std::string element;
		printf("  part/ name: %s", (name = n.GetProperty("name")).c_str());
		try {
			printf(" type: %s", (type = n.GetProperty("type")).c_str());
		} catch (const Exception&) {}
		try {
			printf(" element: %s", (element = n.GetProperty("element")).c_str());
		} catch (const Exception&) {}
		printf("\n");

		if (!element.empty())
		{
			Xml::XmlNode el = GetElement(element);
			ShowElement(el);
		}
		++n;
	}
}


void Wsdl::ShowElement(const Xml::XmlNode& element_node)
{
	std::string name;
	std::string ref;
	try {
		name = element_node.GetProperty("name");
	} catch (const Exception&) {}
	try {
		ref = element_node.GetProperty("ref");
	} catch (const Exception&) {}
	Xml::XmlNode complex(element_node, "complexType");
	if (complex)
	{
		ShowComplexType(complex);
		return;
	}
	else
	if (!ref.empty())
	{
		Xml::XmlNode n = GetElement( ref );
		if (n)
			ShowElement( n );
		return;
	}
	std::string type = element_node.GetProperty("type");
	if (!type.empty())
	{
		printf("    %s  %s\n", name.c_str(), type.c_str());
		try
		{
			Xml::XmlNode n = FindComplexType(type);
			if (n)
			{
				ShowComplexType(n);
			}
		}
		catch (const Exception& ) {}
		return;
	}
	throw WsdlException("no complexType found for element: " + element_node.GetProperty("name"));
}


void Wsdl::ShowComplexType(const Xml::XmlNode& complex)
{
	Xml::XmlNode n(complex, "sequence");
	if (n)
	{
		Xml::XmlNode n(n, "element");
		while (n)
		{
			ShowElement( n );
			++n;
		}
	}
}


Xml::XmlNode Wsdl::GetSchema(const std::string& target_ns) const
{
	Xml::XmlNode root(m_doc);
	if (root.GetNodeName() == "schema" &&
		root.GetProperty("targetNamespace") == target_ns)
		return root;
	Xml::XmlNode n(m_doc, "types");
	if (n)
	{
		Xml::XmlNode n(n, "schema");
		while (n)
		{
			if (n.GetProperty("targetNamespace") == target_ns)
			{
				return n;
			}
			++n;
		}
	}
	throw WsdlException("schema for target namespace not found: " + target_ns);
}


Xml::XmlNode Wsdl::GetElement(const std::string& name) const
{
	std::string ns_prefix = Split(name).NS;
	Xml::XmlNode root(m_doc);
	std::string ns_href = root.GetNsMap()[ns_prefix];
	Xml::XmlNode n = GetSchema(ns_href);
	if (n)
	{
		return GetElement(n, name);
	}
	throw WsdlException("element not found: " + name);
}


Xml::XmlNode Wsdl::FindComplexType(const std::string& type) const
{
	std::string ns_prefix = Split(type).NS;
	std::map<std::string, std::map<std::string, Xml::XmlNode> >::const_iterator it = m_complexTypeMap.find(ns_prefix);
	if (it == m_complexTypeMap.end())
		throw WsdlException("complexType namespace prefix not found: " + ns_prefix);
	const std::map<std::string, Xml::XmlNode>& ref = it -> second;
/*
	{
		for (std::map<std::string, Xml::XmlNode>::const_iterator it = ref.begin(); it != ref.end(); it++)
			printf(" %s %s\n", ns_prefix.c_str(), it -> first.c_str());
	}
*/
	std::map<std::string, Xml::XmlNode>::const_iterator it2 = ref.find(Split(type).Type);
	if (it2 == ref.end())
		throw WsdlException("Can't find complexType: " + type);
	return it2 -> second;
}


Xml::XmlNode Wsdl::FindSimpleType(const std::string& type) const
{
	std::string ns_prefix = Split(type).NS;
	std::map<std::string, std::map<std::string, Xml::XmlNode> >::const_iterator it = m_simpleTypeMap.find(ns_prefix);
	if (it == m_simpleTypeMap.end())
		throw WsdlException("simpleType namespace prefix not found: " + ns_prefix);
	const std::map<std::string, Xml::XmlNode>& ref = it -> second;
/*
	{
		for (std::map<std::string, Xml::XmlNode>::const_iterator it = ref.begin(); it != ref.end(); it++)
			printf(" %s %s\n", ns_prefix.c_str(), it -> first.c_str());
	}
*/
	std::map<std::string, Xml::XmlNode>::const_iterator it2 = ref.find(Split(type).Type);
	if (it2 == ref.end())
		throw WsdlException("Can't find simpleType: " + type);
	return it2 -> second;
}


Xml::XmlNode Wsdl::GetElement(const Xml::XmlNode& node, const std::string& name) const
{
	{
		Xml::XmlNode n(node, "import");
		while (n)
		{
			std::string ns = n.GetProperty("namespace");
			try
			{
				Xml::XmlNode ns_node = GetSchema(ns);
				return GetElement(ns_node, name);
			}
			catch (const WsdlException&)
			{
			}
			++n;
		}
	}
	{
		Xml::XmlNode n(node, "element");
		while (n)
		{
			if (n.GetProperty("name") == Split(name).Type)
			{
				return n;
			}
			++n;
		}
	}
	throw WsdlException("element not found: " + name);
}


/*
Xml::XmlNode Wsdl::FindComplexType(const Xml::XmlNode& node, const std::string& type) const
{
	{
		Xml::XmlNode n(node, "import");
		while (n)
		{
			std::string ns = n.GetProperty("namespace");
			try
			{
				Xml::XmlNode ns_node = GetSchema(ns);
				return FindComplexType(ns_node, type);
			}
			catch (const WsdlException&)
			{
			}
			++n;
		}
	}
	{
		Xml::XmlNode n(node, "complexType");
		while (n)
		{
			std::string name;
			try {
				name = n.GetProperty("name");
			} catch (const Exception&) {}
			if (name == Split(type).Type)
				return n;
			++n;
		}
	}
	throw WsdlException("Complex type not found: " + type);
}
*/


std::list<std::pair<std::string, Xml::XmlNode> > Wsdl::GetMessageParameters(const std::string& msgname, std::string& path) const
{
	std::list<std::pair<std::string, Xml::XmlNode> > vec;
	Xml::XmlNode message = GetMessage(msgname);
	Xml::XmlNode n(message, "part");
	std::string element;

	// part.element --> element
	// part.type --> complex/simple/builtin
	// element.ref --> element
	// element.type --> complex/simple/builtin

	path = "";

	while (n)
	{
		std::string name = n.GetProperty("name");
		if (n.PropertyExists("type"))
		{
			std::string type;
			vec.push_back(std::pair<std::string, Xml::XmlNode>(name, n));
		}
		if (n.PropertyExists("element"))
		{
			element = n.GetProperty("element");
		}
		++n;
	}

	Xml::XmlNode el(m_doc);

	// check 
	if (vec.size() == 1)
	{
		std::list<std::pair<std::string, Xml::XmlNode> >::iterator it = vec.begin();
		Xml::XmlNode& ref = it -> second;
		std::string type = ref.GetProperty("type");
		try
		{
			Xml::XmlNode n = FindComplexType(type);
			vec.erase(it);
			el = n;
		} catch (const Exception&) {}
	}

	if (!element.empty()) // go deeper
	{
		el = GetElement(element);
	}
	if (el)
	{
		while (el)
		{
DEB(printf("/// %s  <%s>\n", el.GetProperty("name").c_str(), el.GetNodeName().c_str());)
			if (el.PropertyExists("name"))
			{
				if (!path.empty())
					path += "/";
				path += el.GetProperty("name");
			}
			if (el.GetNodeName() == "element")
			{
				if (el.PropertyExists("ref"))
				{
					std::string ref = el.GetProperty("ref");
					el = GetElement( ref );
					continue;
				}
				if (el.PropertyExists("type"))
				{
					std::string type = el.GetProperty("type");
					el = FindComplexType(type);
					continue;
				}
				Xml::XmlNode n(el, "complexType");
				if (n)
				{
					ComplexType cc(*this, n);
					std::string ref;
					if (cc.NumberOfMembers() == 1 && cc.GetRef(ref))
					{
						el = GetElement( ref );
						continue;
					}
					vec = cc.GetMembers();
				}
				break;
			}
			else
			if (el.GetNodeName() == "complexType")
			{
				ComplexType cc(*this, el);
				std::string ref;
				if (cc.NumberOfMembers() == 1 && cc.GetRef(ref))
				{
					el = GetElement( ref );
					continue;
				}
				vec = cc.GetMembers();
				break;
			}
			else
			{
				throw WsdlException("Handle type <" + el.GetNodeName() + ">");
			}
		}
	}
	return vec;
}


void Wsdl::ParseElements()
{
	Xml::XmlNode root(m_doc);
	// schema?
	if (root.GetNodeName() == "schema")
	{
		std::string ns_href = root.GetProperty("targetNamespace");
		std::string ns_prefix = root.GetNsMapRe()[ns_href];
		ParseTypes(root, ns_prefix);
		ParseElements(root, ns_prefix);
	}
	// wsdl schema(s)
	{
		Xml::XmlNode n(m_doc, "types");
		if (n)
		{
			Xml::XmlNode n(n, "schema");
			while (n)
			{
				std::string ns_href = n.GetProperty("targetNamespace");
				std::string ns_prefix = root.GetNsMapRe()[ns_href];
				ParseTypes(n, ns_prefix);
				ParseElements(n, ns_prefix);
				//
				++n;
			}
		}
	}
}


void Wsdl::ParseTypes(const Xml::XmlNode& node, const std::string& ns_prefix)
{
	{
		Xml::XmlNode n(node, "complexType");
		while (n)
		{
			std::string name;
			name = n.GetProperty("name");
			m_complexType[ns_prefix].push_back(name);
			m_complexTypeMap[ns_prefix][name] = n;
			//
			ParseComplexType(n, ns_prefix, name);
			++n;
		}
	}

	{
		Xml::XmlNode n(node, "simpleType");
		while (n)
		{
			std::string name;
			name = n.GetProperty("name");
			m_simpleType[ns_prefix].push_back(name);
			m_simpleTypeMap[ns_prefix][name] = n;
			++n;
		}
	}

}


void Wsdl::ParseElements(const Xml::XmlNode& node, const std::string& ns_prefix, const std::string& parent_class)
{
	Xml::XmlNode n(node, "element");
	while (n)
	{
		std::string name;
		if (n.PropertyExists("name"))
			name = n.GetProperty("name");
		std::string type;
		if (n.PropertyExists("type"))
		{
			type = n.GetProperty("type");
		}
		else
		{
			type = /*ns_prefix + ":" +*/ name + "_t";
			// test complex- / simpleType
			Xml::XmlNode complex(n, "complexType");
			if (complex)
			{
				m_complexType[ns_prefix].push_back(type);
				m_complexTypeMap[ns_prefix][type] = complex;
				m_parent[ns_prefix + ":" + type] = parent_class;
				//
				ParseComplexType(complex, ns_prefix, type);
			}
			Xml::XmlNode simple(n, "simpleType");
			if (simple)
			{
				m_simpleType[ns_prefix].push_back(type);
				m_simpleTypeMap[ns_prefix][type] = simple;
				m_parent[ns_prefix + ":" + type] = parent_class;
			}
		}
		// map element name to type
		if (parent_class.empty())
			m_element[ns_prefix][name] = type;
		++n;
	}
}


void Wsdl::ParseComplexType(const Xml::XmlNode& node, const std::string& ns_prefix, const std::string& name)
{
	ComplexType t(*this, node, name, ns_prefix);
	if (t.IsSequence() || t.IsAll())
	{
		Xml::XmlNode n(node, t.IsSequence() ? "sequence" : "all");
		if (n)
		{
			ParseElements(n, ns_prefix, name);
		}
		{
			Xml::XmlNode n(node, "attribute");
			while (n)
			{
				std::string aname = n.GetProperty("name");
				if (n.PropertyExists("type"))
				{
				}
				else
				{
					Xml::XmlNode n(n, "simpleType");
					if (n)
					{
						std::string type = aname + "_t";
						m_simpleType[ns_prefix].push_back(type);
						m_simpleTypeMap[ns_prefix][type] = n;
						m_parent[ns_prefix + ":" + type] = name;
					}
				}
				++n;
			}
		}
	}
}


void Wsdl::CreateInterface()
{
	Xml::XmlNode root(m_doc);

DEB(	Debug deb("Wsdl::CreateInterface");)
	std::string ns = m_appl.Namespace();
	std::string cn = m_appl.Classname();

	std::string h_filename = "I" + cn + ".h";
	std::string cpp_filename = "I" + cn + ".cpp";

	FILE *fil = m_appl.myfopen(h_filename, "wt");
	{
		fprintf(fil, "#include <string>\n");
		fprintf(fil, "#include <vector>\n");
		fprintf(fil, "\n");
		fprintf(fil, "namespace Xml { class Xml::XmlNode; }\n");
		fprintf(fil, "\n");

		NSFil mns(fil, ns); // global namespace

		// simpleType
		for (std::map<std::string, std::list<std::string> >::iterator it = m_simpleType.begin(); it != m_simpleType.end(); it++)
		{
			std::string ns_prefix = it -> first;
			std::list<std::string>& ref = it -> second;
			if (!ref.empty())
			{
				std::string href = root.GetNsMap()[ns_prefix];
				NSFil mns(fil, GetNamespace(ns_prefix), href);
				for (std::list<std::string>::iterator it = ref.begin(); it != ref.end(); it++)
				{
					std::string name = *it;
					std::string parent = m_parent[ns_prefix + ":" + name];
					if (parent.empty())
					{
						SimpleType t(*this, m_simpleTypeMap[ns_prefix][name], name, ns_prefix);
						t.CreateInterface(fil);
					}
				}
				fprintf(fil, "\n");
			}
		}

		// create class definitions
		bool repeat = false;
		bool gen = false;
		do
		{
			repeat = false;
			gen = false;
			for (std::map<std::string, std::list<std::string> >::iterator it = m_complexType.begin(); it != m_complexType.end(); it++)
			{
				std::string ns_prefix = it -> first;
				std::list<std::string>& ref = it -> second;
				if (!ref.empty())
				{
					std::string href = root.GetNsMap()[ns_prefix];
					NSFil mns(fil, GetNamespace(ns_prefix), href);
					for (std::list<std::string>::iterator it = ref.begin(); it != ref.end(); it++)
					{
						std::string name = *it;
						std::string fqn = ns_prefix + ":" + name;
						std::string parent = m_parent[fqn];
						if (parent.empty())
						{
						ComplexType t(*this, m_complexTypeMap[ns_prefix][name], name, ns_prefix);
						if (!t.IsGenerated())
						{
							if (t.OkToGenerate())
							{
								std::string doc = t.Documentation();
								fprintf(fil, "\t// -----------------------------------------------------------------------------\n");
								if (!doc.empty())
								{
									fprintf(fil, "\t/** %s */\n", doc.c_str());
								}
								t.CreateInterface(fil);
								t.SetGenerated();
								gen = true;
							}
							else
							{
								repeat = true;
							}
						}
						}
					}
				}
			}
		}
		while (repeat && gen);

		//
		fprintf(fil, "\n\n");
	}

	// close .h file
	fclose(fil);

}


void Wsdl::CreateImplementation()
{
	Xml::XmlNode root(m_doc);

DEB(	Debug deb("Wsdl::CreateInterface");)
	std::string ns = m_appl.Namespace();
	std::string cn = m_appl.Classname();

	std::string h_filename = "I" + cn + ".h";
	std::string cpp_filename = "I" + cn + ".cpp";

	FILE *fil = m_appl.myfopen(cpp_filename, "wt");
	{
		fprintf(fil, "#include \"%s\"\n", h_filename.c_str());
		fprintf(fil, "#include <Sockets/XmlNode.h>\n");
		fprintf(fil, "#include <Sockets/XmlException.h>\n");
		fprintf(fil, "\n");

		NSFil mns(fil, ns); // global namespace

		// create class implementations
		for (std::map<std::string, std::list<std::string> >::iterator it = m_complexType.begin(); it != m_complexType.end(); it++)
		{
			std::string ns_prefix = it -> first;
			std::list<std::string>& ref = it -> second;
			if (!ref.empty())
			{
				std::string href = root.GetNsMap()[ns_prefix];
				NSFil mns(fil, GetNamespace(ns_prefix), href);
				for (std::list<std::string>::iterator it = ref.begin(); it != ref.end(); it++)
				{
					std::string name = *it;
					ComplexType t(*this, m_complexTypeMap[ns_prefix][name], name, ns_prefix);
					t.CreateImplementation(fil);
				}
			}
		}

	}
	fclose(fil);
}


const std::string Wsdl::GetParent(const std::string& classname) const
{
	std::map<std::string, std::string>::const_iterator it;
	if ( (it = m_parent.find(classname)) != m_parent.end())
	{
		return it -> second;
	}
	return "";
}


const std::list<std::string>& Wsdl::ComplexTypes(const std::string& ns_prefix) const
{
	std::map<std::string, std::list<std::string> >::const_iterator it = m_complexType.find(ns_prefix);
	if (it != m_complexType.end())
		return it -> second;
	throw WsdlException("no complexTypes in ns: " + ns_prefix);
}


const std::list<std::string>& Wsdl::SimpleTypes(const std::string& ns_prefix) const
{
	std::map<std::string, std::list<std::string> >::const_iterator it = m_simpleType.find(ns_prefix);
	if (it != m_simpleType.end())
		return it -> second;
	throw WsdlException("no simpleTypes in ns: " + ns_prefix);
}


const std::string Wsdl::NamespaceFromHref(const std::string& href) const
{
	size_t pos = href.find("://");
	if (pos != std::string::npos)
	{
		std::string hreft = href.substr(pos + 3);
		pos = hreft.find("/");
		std::string host = hreft;
		std::string path;
		if (pos != std::string::npos)
		{
			host = hreft.substr(0, pos);
			path = hreft.substr(pos + 1);
		}
		Parse pa(host, ".:");
		std::string tmp = pa.getword();
		std::string ns;
		std::string ns2;
		bool first = true;
		while (!tmp.empty())
		{
			if (first)
				first = false;
			else
			{
				ns = ns2;
				ns2 = tmp + "_" + ns2;
			}
			tmp = pa.getword();
		}
		for (size_t i = 0; i < path.size(); i++)
			if (isalnum(path[i]))
				ns2 += path[i];
			else
				ns2 += "_";
		return ns2;
	}
	std::string ns;
	for (size_t i = 0; i < href.size(); i++)
		if (!isalnum(href[i]))
			ns += "_";
		else
			ns += href[i];
	return ns;
}


const std::string Wsdl::NamespaceFromPrefix(const std::string& prefix) const
{
	Xml::XmlNode root(m_doc);
	std::string href = root.GetNsMap()[prefix];
	return NamespaceFromHref(href);
}


const std::string Wsdl::GetNamespace(const std::string& prefix) const
{
	Xml::XmlNode root(m_doc);
	std::string href = root.GetNsMap()[prefix];
	if (m_appl.Prefix())
	{
		return root.GetNsMapRe()[href];
	}
	return NamespaceFromHref(href);
}


const std::string Wsdl::GetRealPrefix(const std::string& prefix) const
{
	Xml::XmlNode root(m_doc);
	std::string href = root.GetNsMap()[prefix];
	return root.GetNsMapRe()[href];
}


const std::string Wsdl::GetRealType(const std::string& type) const
{
	Xml::XmlNode root(m_doc);
	std::string ns = Split(type).NS;
	std::string t = Split(type).Type;
	std::string href = root.GetNsMap()[ns];
	return root.GetNsMapRe()[href] + ":" + t;
}


